/*
 * IdealIRC - Internet Relay Chat client
 * Copyright (C) 2019  Tom-Andre Barstad
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include "ServerItem.h"
#include <QDebug>

ServerItem::ServerItem()
	: m_type(Type::Root)
{}

ServerItem::ServerItem(ServerItem* parentItem, const QString& serverName, const QString& host, const quint16 port, const QString& password)
	: m_type(Type::Server)
	, m_serverName(serverName)
	, m_host(host)
	, m_port(port)
	, m_password(password)
	, m_parent(parentItem)
{}

ServerItem::ServerItem(const QString& networkName, const QString& host, const quint16 port, const QString& password)
	: m_type(Type::Network)
	, m_networkName(networkName)
	, m_host(host)
	, m_port(port)
	, m_password(password)
{}

ServerItem::~ServerItem()
{
	qDeleteAll(m_children);
}

ServerItem::Type ServerItem::getType() const
{
	return m_type;
}

const QString& ServerItem::getNetworkName() const
{
	return m_networkName;
}

const QString& ServerItem::getServerName() const
{
	return m_serverName;
}

const QString& ServerItem::getHost() const
{
	return m_host;
}

quint16 ServerItem::getPort() const
{
	return m_port;
}

const QString& ServerItem::getPassword() const
{
	return m_password;
}

ServerItem*ServerItem::getParentItem() const
{
	return m_parent;
}

bool ServerItem::isOrphan() const
{
	return m_parent == nullptr;
}

bool ServerItem::addChild(ServerItem* child)
{
	if (m_type == Type::Server) {
		qWarning() << "Tried to add a child item on a server entry!";
		return false;
	}
	m_children.push_back(child);
	return true;
}

bool ServerItem::delChild(ServerItem* child)
{
	if (m_type == Type::Server) {
		qWarning() << "Tried to delete a child item on a server entry!";
		return false;
	}
	m_children.removeOne(child);
	return true;
}

ServerItem* ServerItem::getChild(int row) const
{
	if (m_type == Type::Server) {
		qWarning() << "Tried to get a child item on a server entry!";
		return nullptr;
	}

	if (row >= m_children.count()) {
		qWarning() << "Tried to get child from row" << row << "- but it's out of index. Network:" << m_networkName;
		return nullptr;
	}

	return m_children[row];
}

int ServerItem::childCount() const
{
	return m_children.count();
}

int ServerItem::childRow(ServerItem* child) const
{
	return m_children.indexOf(child);
}

ServerItem*ServerItem::getParent() const
{
	return m_parent;
}
