/*
 * IdealIRC - Internet Relay Chat client
 * Copyright (C) 2019  Tom-Andre Barstad
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include "ServerModel.h"
#include "config.h"
#include <QDebug>

ServerModel::ServerModel(QObject *parent)
	: QAbstractItemModel(parent)
{
	IniFile ini(LOCAL_PATH+"/servers.ini");
	buildToplevelItem(ini, "NONE"); // Orphaned servers. Always build first.
	const int sectionCount = ini.countSections();
	for (int i = 0; i < sectionCount; ++i) {
		QString section = ini.section(i);
		if (section == "NONE")
			continue;

		buildToplevelItem(ini, section);
	}

	qDebug() << "Built server tree model.";
}

ServerModel::~ServerModel()
{
	qDeleteAll(toplevelItems);
}

ServerModel::ItemIniValue::ItemIniValue(QString value)
{
	if (value.isEmpty())
		return;

	// Value format: host:port|password
	if (value.contains('|')) {
		password = value.split('|')[1];
		value = value.remove("|"+password);
	}
	if (value.contains(':')) {
		port = value.split(':')[1];
		value = value.remove(":"+port);
	}
	host = value;
}

QModelIndex ServerModel::index(int row, int column, const QModelIndex &parent) const
{
	// TEST
	//if (!hasIndex(row, column, parent))
	//	return QModelIndex();

	ServerItem* item{ nullptr };
	if (!parent.isValid()) {
		if (row < toplevelItems.count())
			item = toplevelItems[row];
	}
	else
		item = static_cast<ServerItem*>(parent.internalPointer());

	if (item)
		return createIndex(row, column, item);
	else
		return QModelIndex();
}

QModelIndex ServerModel::parent(const QModelIndex &index) const
{
	// FIXME: Implement me!
	if (!index.isValid())
		return QModelIndex();

	ServerItem* childItem = static_cast<ServerItem*>(index.internalPointer());
	if (!childItem->isOrphan()) {
		ServerItem* parentItem = childItem->getParentItem();
		return createIndex(parentItem->childRow(childItem), 0, parentItem);
	}
	else {
		return QModelIndex();
	}
}

int ServerModel::rowCount(const QModelIndex &parent) const
{
	if (!parent.isValid())
		return toplevelItems.count();

	ServerItem* item = static_cast<ServerItem*>(parent.internalPointer());
	if (!item)
		return 0;
	else
		return item->childCount();
}

int ServerModel::columnCount(const QModelIndex&) const
{
	return 1;
}

QVariant ServerModel::data(const QModelIndex &index, int role) const
{
	if (!index.isValid())
		return QVariant();

	// FIXME: Implement me!
	return "neger";
}

bool ServerModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
	if (data(index, role) != value) {
		// FIXME: Implement me!
		emit dataChanged(index, index, QVector<int>() << role);
		return true;
	}
	return false;
}

Qt::ItemFlags ServerModel::flags(const QModelIndex &index) const
{
	if (!index.isValid())
		return Qt::NoItemFlags;

	return Qt::ItemIsEditable; // FIXME: Implement me!
}

bool ServerModel::insertRows(int row, int count, const QModelIndex &parent)
{
	beginInsertRows(parent, row, row + count - 1);
	// FIXME: Implement me!
	endInsertRows();
}

bool ServerModel::removeRows(int row, int count, const QModelIndex &parent)
{
	beginRemoveRows(parent, row, row + count - 1);
	// FIXME: Implement me!
	endRemoveRows();
}

void ServerModel::buildToplevelItem(IniFile& ini, const QString& section)
{
	if (!ini.sectionExists(section))
		return;

	ServerItem* parent{ nullptr };
	if (section != "NONE") {
		ItemIniValue networkDefault(ini.value(section, "DEFAULT"));
		parent = new ServerItem(section, networkDefault.host,
											networkDefault.port.toUShort(),
											networkDefault.password);
		toplevelItems.push_back(parent);
	}

	const int itemCount = ini.countItems(section);
	for (int i = 0; i < itemCount; ++i) {
		QString key = ini.key(section, i);
		if (key == "DEFAULT")
			continue;

		ItemIniValue val(ini.value(section, key));
		ServerItem* item = new ServerItem(parent, key, val.host, val.port.toUShort(), val.password);
		if (parent)
			parent->addChild(item);
		else
			toplevelItems.push_back(item);
	}
}
