#include "DCC.h"
#include "IRCBase.h"
#include <fmt/format.h>
#include <iostream>

using asio::ip::tcp;

struct DCCPriv
{
	DCCPriv(DCC& super_, IRCBase& ircctx_, DCC::Direction dir_, asio::io_context& ioctx_, const std::string& ip_, const std::string& port_)
		: super(super_)
		, ircctx(ircctx_)
		, direction(dir_)
		, ioctx(ioctx_)
		, ip(ip_)
		, port(port_)
		, resolver(ioctx_)
		, acceptor(ioctx_)
		, endpoint(tcp::v4(), std::stoi(port_))
		, socket(ioctx_)
	{}

	DCC& super;
	IRCBase& ircctx;

	bool pending{ true };
	const DCC::Direction direction;
	asio::io_context& ioctx;
	std::string ip; // Only used when Direction=Target, and describes the Initiator's IP.
	std::string port;

	tcp::resolver resolver; // for outbound connection (we act as a Target)
	tcp::acceptor acceptor; // for inbound connection (we act as an Initiator)

	tcp::resolver::results_type resolverResults;
	asio::ip::tcp::endpoint endpoint;
	tcp::socket socket;

	asio::streambuf readbuf;

	DCC::CallbackRead cbRead;
	DCC::CallbackConnected cbCon;
	DCC::CallbackDisconnected cbDiscon;

	void connected(const asio::error_code& ec)
	{
		super.onConnected();

		asio::streambuf::mutable_buffers_type mutableBuf{ readbuf.prepare(16384) };
		socket.async_receive(mutableBuf,
			[this](const asio::error_code& ec, std::size_t rl){
				readbuf.commit(rl);
				read(ec);
			});
	}

	void disconnected(const asio::error_code& ec)
	{
		super.onDisconnected(SystemErrorToIRCError(ec));
	}

	void read(const asio::error_code& ec)
	{
		if (ec) {
			disconnected(ec);
			return;
		}

		std::istream is(&readbuf);
		super.onRead(is);

		asio::streambuf::mutable_buffers_type mutableBuf{ readbuf.prepare(16384) };
		socket.async_receive(mutableBuf,
							 [this](const asio::error_code& ec, std::size_t rl){
								 readbuf.commit(rl);
								 read(ec);
							 });
	}

	void writeFinished(const asio::error_code& ec, std::size_t bytes_transferred)
	{
		std::cout << "DCC wrote " << bytes_transferred << " bytes." << std::endl;
		if (!ec) return; // No errors.
		std::cerr << fmt::format("'- DCC write error: {} ({})", SystemErrorToIRCError(ec), ec.message()) << std::endl;
	}
};

DCC::DCC(IRCBase& ircctx, asio::io_context& ioctx, const std::string& port)
		: mp(std::make_unique<DCCPriv>(*this, ircctx, Direction::Initiator, ioctx, "", port))
{
	mp->acceptor.open(mp->endpoint.protocol());
	mp->acceptor.bind(mp->endpoint);
	mp->acceptor.listen(1);
	mp->acceptor.async_accept(mp->socket, [this](const asio::error_code& ec){
		if (ec)
			mp->disconnected(ec);
		else
			mp->connected(ec);
	});
}

DCC::DCC(IRCBase& ircctx, asio::io_context& ioctx, const std::string& ip, const std::string& port)
	: mp(std::make_unique<DCCPriv>(*this, ircctx, Direction::Target, ioctx, ip, port))
{}

DCC::DCC(DCC&& other) noexcept
	: mp(std::move(other.mp))
{}

DCC::~DCC()
{
	try {
		if (mp->socket.is_open()) {
			mp->socket.shutdown(asio::socket_base::shutdown_both);
			mp->socket.close();
		}
		if (mp->direction == Direction::Initiator)
			mp->acceptor.close();
	}
	catch(const asio::system_error& e) {
		// Properly log e.what() or e.code() maybe?
		std::cerr << fmt::format("~DCC() caught exception ({}): {}\n       IP={}, Port={}, Direction={}",
									e.code().value(), e.what(), mp->ip, mp->port, mp->direction) << std::endl;
	}
}

bool DCC::isPending() const
{
	return mp->pending;
}

IRCError DCC::accept()
{
	if (mp->direction == Direction::Initiator)
		return IRCError::DCC_NotATarget; // Only usable if we are a Target...

	mp->resolverResults = mp->resolver.resolve(mp->ip, mp->port);

	asio::async_connect(mp->socket, mp->resolverResults,
		[this](const asio::error_code& ec, const tcp::endpoint&) {
			if (ec)
				mp->disconnected(ec);
			else
				mp->connected(ec);
		});

	mp->pending = false;
	return IRCError::NoError;
}

DCC::Direction DCC::direction() const
{
	return mp->direction;
}

void DCC::write(const ByteString& data)
{
	mp->socket.async_send(asio::buffer(data),
		[this](const asio::error_code& ec, std::size_t bytes_transferred){
			mp->writeFinished(ec, bytes_transferred);
		});
}

IRCBase& DCC::context()
{
	return mp->ircctx;
}

const IRCBase& DCC::context() const
{
	return mp->ircctx;
}

void DCC::callbackRead(DCC::CallbackRead&& cb)
{
	mp->cbRead = std::move(cb);
}

void DCC::callbackConnected(DCC::CallbackConnected&& cb)
{
	mp->cbCon = std::move(cb);
}

void DCC::callbackDisconnected(DCC::CallbackDisconnected&& cb)
{
	mp->cbDiscon = std::move(cb);
}


void DCC::onRead(std::istream& is)
{
	if (mp->cbRead)
		mp->cbRead(is);
}

void DCC::onConnected()
{
	if (mp->cbCon)
		mp->cbCon();
}

void DCC::onDisconnected(IRCError e)
{
	if (mp->cbDiscon)
		mp->cbDiscon(e);
}

