#ifndef DCC_H
#define DCC_H

#include "IRCError.h"
#include <string>
#include <memory>
#include <asio.hpp>

using asio::ip::tcp;

class IRCBase;

struct DCCPriv;
class DCC
{
	friend struct IRCBasePriv;
	friend struct DCCPriv;

public:
	using CallbackRead = std::function<void(std::istream&)>;
	using CallbackConnected = std::function<void()>;
	using CallbackDisconnected = std::function<void(IRCError e)>;
	using ByteString = std::basic_string<std::byte>;

	/*
	 * Initiator acts as the server, and the target is the client
	*/
	enum class Direction {
		Initiator,
		Target
	};

	~DCC();
	DCC(DCC&& other) noexcept;

	//! Constructs an Initiator (server)
	DCC(IRCBase& ircctx, asio::io_context& ioctx, const std::string& port);

	//! Constructs a Target (client)
	DCC(IRCBase& ircctx, asio::io_context& ioctx, const std::string& ip, const std::string& port);

	[[nodiscard]] bool isPending() const;
	IRCError accept();

	[[nodiscard]] Direction direction() const;

	void write(const ByteString& data);

	IRCBase& context();
	const IRCBase& context() const;

	/*
	 * Callbacks used by 'users' of this class.
	*/
	void callbackRead(CallbackRead&& cb);
	void callbackConnected(CallbackConnected&& cb);
	void callbackDisconnected(CallbackDisconnected&& cb);

protected:
	/*
	 * Use these to read data in a custom derivative of DCC.
	 * If you re-implement any of these, make sure to forward the
	 * calls to this base class at the end of your function.
	*/
	virtual void onRead(std::istream& is);
	virtual void onConnected();
	virtual void onDisconnected(IRCError e);

private:
		std::unique_ptr<DCCPriv> mp;
};

#endif // DCC_H
