#include "Utilities.h"
#include <fmt/format.h>
#include <sstream>

std::pair<std::string,std::string> FormatCTCPLine(std::string line)
{
	/* Clean away the CTCP byte flags */
	if (line[0] == CTCPflag)
		line.erase(line.begin());

	auto endIt = std::find(line.begin(), line.end(), CTCPflag);
	if (endIt != line.end())
		line.pop_back();

	auto cmdIt = std::find(line.begin(), line.end(), ' ');
	std::string cmd(line.begin(), cmdIt);
	std::string ctcpMsg;
	if (cmdIt != line.end())
		ctcpMsg = std::string(cmdIt + 1, line.end());

	return std::make_pair(cmd, ctcpMsg);
}

/*!
 * @brief Simple wildcard match, supporting only a single wildcard.
 * @returns true if string 'str' matches with the wildcard string 'wc'.
*/
bool singleWildcardMatch(const std::string& str, const std::string& wc)
{
	if (str.empty() || wc.empty())
		return false;

	auto wcit = std::find(wc.begin(), wc.end(), '*');
	if (wcit == wc.end())
		return str == wc;

	/*
	 * Since we only support a single wildcard we only need to match the left side and right side
	 * of the asterisk (wildcard) against the target string 'str'.
	 * Also if the "end" of left match is equal to the "begin" of right match, it's a mismatch.
	 */

	bool hasLeft, hasRight { false };
	auto leftEnd = str.end();
	auto rightBegin = str.end();

	/* Left side */
	if (wcit != wc.begin()) {
		std::string wildls(wc.begin(), wcit);
		auto strmatch = str.substr(0, wildls.length());
		if (wildls != strmatch)
			return false;
		leftEnd = str.begin() + wildls.length();
		hasLeft = true;
	}

	/* Right side */
	if (wcit != wc.end() - 1) {
		std::string wildrs(wcit + 1, wc.end());
		auto mpos = str.find(wildrs);
		if (mpos == std::string::npos)
			return false;
		auto strmatch = str.substr(mpos);
		if (wildrs != strmatch)
			return false;
		rightBegin = str.begin() + mpos;
		hasRight = true;
	}

	// Avoid matching "left*right" with "leftright"
	return !(hasLeft && hasRight && leftEnd == rightBegin);
}

std::string longIpToNormal(const std::string& longip)
{
	try {
		unsigned ip = std::stoul(longip);
		unsigned a = (ip & 0xFF000000u) >> 24u;
		unsigned b = (ip & 0x00FF0000u) >> 16u;
		unsigned c = (ip & 0x0000FF00u) >> 8u;
		unsigned d = ip & 0x000000FFu;
		return fmt::format("{}.{}.{}.{}", a, b, c, d);
	}
	catch (...) {
		return "0.0.0.0";
	}
}

std::string normalIpToLong(const std::string& normalip)
{
	std::stringstream ss(normalip);
	std::string p;
	std::vector<unsigned> parts;
	while (getline(ss, p, '.'))
		parts.emplace_back(std::stoul(p));

	uint32_t ret = parts[0];
	for (int i = 1; i < 4; ++i) {
		ret <<= 8u;
		ret |= parts[i];
	}

	return std::to_string(ret);
}
